unit Main;

{ Wolfs Sudoku-Löser.
  Geschrieben von Wolf Behrenhoff <wolf.behrenhoff@gmx.net>, 2006.
  Dieses Programm ist freie Software im Sinne der GNU General Public License.
  Siehe http://www.gnu.org/licenses/gpl.html für Details.

  Nach Lazarus umgewandelt und leicht modifiziert.
  }

//{$mode objfpc}{$H+}
{$MODE Delphi}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ComCtrls,
  Menus,
  LCLIntf, LCLType, ExtCtrls;

type
  TFeldStatus=(manuell, automatisch, geraten, geladen);
  TSudokuZahlenmenge=Set of 1..9;
  TSudoku = class
    public
      procedure AddUndo(s: String);
      function CanUndo: Boolean;
      constructor Create;
      destructor Destroy; override;
      function ErlaubtQ(x,y,zahl: Integer): Boolean;
      procedure Finde(var x,y,zahl: Integer);
      function FindeLoesung: Boolean;
      function Geloest: Boolean;
      function GetFindGrund: String;
      function getHint(x,y: Integer): String;
      function GetStatus(x,y: Integer): TFeldStatus;
      function GetUndoString: String;
      function GetZahl(x,y: Integer): Integer;
      procedure LoadFromFile(filename: String);
      procedure LoescheZahl(x,y: Integer);
      procedure SaveToFile(filename: String);
      procedure SetZahl(x,y,zahl: Integer; stat: TFeldStatus);
      procedure Undo;
      function Unloesbar: Boolean;
    private
      Feld: Array[0..8, 0..8] of Byte;
      Status: Array[0..8, 0..8] of TFeldStatus;
      Erlaubt: Array[0..8, 0..8] of TSudokuZahlenmenge;
      UndoList: TStringList;
      grund: String;
      procedure ErmittleAusschlussNeu;
      procedure GetFreiesFeld(var x,y: Integer);
      procedure RemoveInternalUndos;
      procedure SchliesseAusVon(x,y: Integer);
  end;


type

  { TForm1 }

  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    Datei1: TMenuItem;
    ber1: TMenuItem;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    Separator2: TMenuItem;
    Bearbeiten1: TMenuItem;
    Hilfe1: TMenuItem;
    Neu1: TMenuItem;
    Laden1: TMenuItem;
    Speichern1: TMenuItem;
    Beenden1: TMenuItem;
    Rckgngig1: TMenuItem;
    Displayaktualisieren1: TMenuItem;
    Separator1: TMenuItem;
    StatusBar: TStatusBar;
    ZeigeNaechsteBtn: TButton;
    FindSolution: TButton;
    ZeileAlleBtn: TButton;
    EndBtn: TButton;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    procedure ber1Click(Sender: TObject);
    procedure Displayaktualisieren1Click(Sender: TObject);
    procedure EingabeChange(Sender: TObject);
    procedure EndBtnClick(Sender: TObject);
    procedure FindSolutionClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Laden1Click(Sender: TObject);
    procedure Neu1Click(Sender: TObject);
    procedure Rckgngig1Click(Sender: TObject);
    procedure Speichern1Click(Sender: TObject);
    procedure ZeigeNaechsteBtnClick(Sender: TObject);
    procedure ZeileAlleBtnClick(Sender: TObject);
  private
    inp: Array[0..8, 0..8] of TMemo;  // Edit
    s: TSudoku;
    procedure AktualisiereDisplay;
    procedure AktualisiereHintTexte;
    procedure AktualisiereUndo;
    function ZeigeNaechsteZahl: Boolean;

  public

  end;

const
  colors: Array[TFeldStatus] of TColor=(clWhite,clLime,clYellow,clLtGray);

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

{%region /fold 'TForm1'}

procedure TForm1.ber1Click(Sender: TObject);
begin
  MessageDlg('Sudoku-Löser'#10+
    'Autor: Wolf Behrenhoff'#10+
    'Datum: April 2006'#10+
    'Lizenz: GNU General Public License, siehe http://www.gnu.org/licenses/gpl.html'#10+
    #10+
    'Viel Spaß beim Lösen von Sudokus!',
    mtInformation,[mbOK],0);
end;

procedure TForm1.Displayaktualisieren1Click(Sender: TObject);
begin
AktualisiereDisplay;
Invalidate;
end;

procedure TForm1.EingabeChange(Sender: TObject);
var
  str: String;
  x,y: Integer;
begin
str:=(Sender as TMemo).Text;
x:=TMemo(Sender).Tag mod 9;
y:=TMemo(Sender).Tag div 9;
s.AddUndo('Benutzereingabe');
AktualisiereUndo;
StatusBar.SimpleText:='Benutzereingabe';
if s.GetZahl(x,y)>0 then
  begin
  s.LoescheZahl(x,y);
  inp[x,y].Color:=clWhite;
  end;
if (length(str)=1)
  then
    begin
    if (str[1] in ['1'..'9'])and s.ErlaubtQ(x,y,StrToInt(str[1]))
      then
        begin
        s.SetZahl(x,y,StrToInt(str[1]),manuell);
        inp[x,y].Color:=clWhite;
        end
      else
        begin
        inp[x,y].Text:='X';
        inp[x,y].SelStart:=0;
        inp[x,y].SelLength:=1;
        inp[x,y].Color:=clRed;
        StatusBar.SimpleText:='Bitte eine erlaubte Zahl eingeben';
        end;
    end
  else
    begin
    inp[x,y].Color:=clWhite;
    end;
AktualisiereHintTexte;
end;

procedure TForm1.EndBtnClick(Sender: TObject);
begin
Close;
end;

procedure TForm1.FindSolutionClick(Sender: TObject);
begin
s.AddUndo('Brute-Force-Lösung');
if s.FindeLoesung
  then
    begin
    StatusBar.SimpleText:='Lösung gefunden!';
    end
  else
    begin
    MessageDlg('Keine Lösung möglich',mtError,[mbOK],0);
    StatusBar.SimpleText:='Keine Lösung möglich';
    s.Undo;
    end;
AktualisiereDisplay;
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  x, y: Integer;
begin
Application.HintHidePause:=20000;
s:=TSudoku.Create;
for y:=0 to 8 do for x:=0 to 8 do
  begin
  inp[x,y]:=TMemo.Create(self);
  with inp[x,y] do
    begin
    Parent:=Form1;
    Left:=x*35+10+10*(x div 3);
    Top:=y*35+10+10*(y div 3);
    Height:=33;
    Width:=33;
    Tag:=9*y+x;
    OnChange:=EingabeChange;
    ShowHint:=True;
    MaxLength:=1;
    Alignment:=taCenter;
    Font.Size:=15;
    end;
  end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
s.Free;
end;

procedure TForm1.Laden1Click(Sender: TObject);
begin
if OpenDialog1.Execute
  then
    begin
    s.AddUndo('Datei laden');
    s.LoadFromFile(OpenDialog1.Filename);
    AktualisiereDisplay;
    StatusBar.SimpleText:='Datei erfolgreich geladen';
    end
  else StatusBar.SimpleText:='Laden abgebrochen';
end;

procedure TForm1.Neu1Click(Sender: TObject);
var
  x,y: Integer;
begin
s.AddUndo('Neu');
for y:=0 to 8 do for x:=0 to 8 do s.LoescheZahl(x,y);
AktualisiereDisplay;
StatusBar.SimpleText:='Leeres Feld erstellt.';
end;

procedure TForm1.Rckgngig1Click(Sender: TObject);
begin
s.Undo;
AktualisiereDisplay;
StatusBar.SimpleText:='Viel Spaß mit Wolfs Sudoku-Helfer';
end;

procedure TForm1.Speichern1Click(Sender: TObject);
begin
if SaveDialog1.Execute
  then
    begin
    s.SaveToFile(SaveDialog1.Filename);
    StatusBar.SimpleText:='gespeichert.';
    end
  else StatusBar.SimpleText:='nicht gespeichert';
end;

procedure TForm1.ZeigeNaechsteBtnClick(Sender: TObject);
begin
s.AddUndo('Eine einfache Zahl ermitteln');
AktualisiereUndo;
ZeigeNaechsteZahl;
StatusBar.SimpleText:=s.getFindGrund;
end;

procedure TForm1.ZeileAlleBtnClick(Sender: TObject);
begin
s.AddUndo('Alle einfachen Zahlen ermitteln');
AktualisiereUndo;
repeat
until not ZeigeNaechsteZahl;
StatusBar.SimpleText:='Alle "einfachen" Zahlen wurden eingetragen';
end;

procedure TForm1.AktualisiereDisplay;
var
  x,y: Integer;
  old: TNotifyEvent;
begin
for x:=0 to 8 do for y:=0 to 8 do
  begin
  old:=inp[x,y].OnChange;
  inp[x,y].OnChange:=nil;
  if s.GetZahl(x,y)=0
    then
      begin
      inp[x,y].Text:='';
      end
    else
      begin
      inp[x,y].Text:=IntToStr(s.GetZahl(x,y));
      end;
  inp[x,y].Color:=colors[s.GetStatus(x,y)];
  inp[x,y].OnChange:=old;
  end;
AktualisiereHintTexte;
AktualisiereUndo;
end;

procedure TForm1.AktualisiereHintTexte;
var
  x,y: Integer;
begin
for x:=0 to 8 do for y:=0 to 8 do inp[x,y].Hint:=s.getHint(x,y);
end;

procedure TForm1.AktualisiereUndo;
begin
Rckgngig1.Caption:='&Rückgängig ('+s.GetUndoString+')';
Rckgngig1.Enabled:=s.CanUndo;
end;

function TForm1.ZeigeNaechsteZahl: Boolean;
var
  x,y,zahl: Integer;
begin
s.Finde({%H-}x,{%H-}y,{%H-}zahl);
if zahl>0
  then
    begin
    s.SetZahl(x,y,zahl,automatisch);
    //inp[x,y].Text:=IntToStr(zahl);
    //inp[x,y].Color:=clLime;
    AktualisiereHintTexte;
    AktualisiereDisplay;
    Result:=True;
    end
  else Result:=False;
end;

{%endregion TForm1}

{ TSudoku }

{%region /fold 'TSudoku'}
procedure TSudoku.AddUndo(s: String);
var
  p: Pointer;
begin
GetMem(p,SizeOf(Feld)+SizeOf(Status)+SizeOf(Erlaubt));
Move(Feld,p^,SizeOf(Feld));
Move(Status,(p+SizeOf(Feld))^,SizeOf(Status));                                  // Move(Status,Pointer(LongWord(p)+SizeOf(Feld))^,SizeOf(Status));
Move(Erlaubt,(p+SizeOf(Feld)+SizeOf(Status))^,SizeOf(Erlaubt));                 // Move(Erlaubt,Pointer(LongWord(p)+SizeOf(Feld)+SizeOf(Status))^,SizeOf(Erlaubt));
UndoList.Objects[UndoList.Add(s)]:=p;
end;

function TSudoku.CanUndo: Boolean;
begin
Result:=UndoList.Count>0;
end;

constructor TSudoku.Create;
var
  x, y: Integer;
begin
UndoList:=TStringList.Create;
for x:=0 to 8 do for y:=0 to 8 do
  begin
  erlaubt[x,y]:=[1,2,3,4,5,6,7,8,9];
  Feld[x,y]:=0;
  Status[x,y]:=manuell;
  end;
end;

destructor TSudoku.Destroy;
begin
UndoList.Free;
inherited;
end;

function TSudoku.ErlaubtQ(x,y,zahl: Integer): Boolean;
begin
Result:=zahl in erlaubt[x,y];
end;

procedure TSudoku.Finde(var x,y,zahl: Integer);
var
  i,j,z: Integer;
  inSpalte, inZeile, in3x3, k3x, k3y,
  inSx, inSy, inZx, inZy, in3x, in3y: Integer;
begin
zahl:=0;
for z:=1 to 9 do
  begin
  for j:=0 to 8 do
    begin
    inSpalte:=0;
    inZeile:=0;
    in3x3:=0;
    for i:=0 to 8 do
      begin
      //irgendwo nur 1 Zahl erlaubt?
      if (erlaubt[i,j]=[z]) and (Feld[i,j]=0) then
        begin
        x:=i; y:=j; zahl:=z;
        grund:='Auf dem Feld ('+IntToStr(x+1)+','+IntToStr(y+1)+') '+
               'darf ausschließlich die '+IntToStr(z)+' stehen.';
        Exit;
        end;
      //in irgendeiner Spalte eine Zahl nur 1x erlaubt?
      if z in erlaubt[j,i] then
        begin
        Inc(inSpalte);
        inSx:=j; inSy:=i;
        end;
      //in irgendeiner Zeile eine Zahl nur 1x erlaubt?
      if z in erlaubt[i,j] then
        begin
        Inc(inZeile);
        inZx:=i; inZy:=j;
        end;
      //in einem 3x3-Kasten eine Zahl nur 1x erlaubt?
      k3x:=(j mod 3)*3+i mod 3;
      k3y:=(j div 3)*3+i div 3;
      if z in erlaubt[k3x,k3y] then
        begin
        Inc(in3x3);
        in3x:=k3x; in3y:=k3y;
        end;
      end;
    if (inSpalte=1) and (Feld[inSx,inSy]=0) then
      begin
      x:=inSx; y:=inSy; zahl:=z;
      grund:='Auf ('+IntToStr(x+1)+','+IntToStr(y+1)+') '+
             'ist die einzige Position für die '+IntToStr(z)+' in der Spalte.';
      Exit;
      end;
    if (inZeile=1) and (Feld[inZx,inZy]=0) then
      begin
      x:=inZx; y:=inZy; zahl:=z;
      grund:='Auf ('+IntToStr(x+1)+','+IntToStr(y+1)+') '+
             'ist die einzige Position für die '+IntToStr(z)+' in der Zeile.';
      Exit;
      end;
    if (in3x3=1) and (Feld[in3x,in3y]=0) then
      begin
      x:=in3x; y:=in3y; zahl:=z;
      grund:='Auf ('+IntToStr(x+1)+','+IntToStr(y+1)+') '+
             'ist die einzige Position für die '+IntToStr(z)+' im 3x3-Feld.';
      Exit;
      end;
    end;
  end;
grund:='Keine "einfache" Zahl gefunden';
end;

function TSudoku.FindeLoesung: Boolean;

  procedure EinfacheFinden;
  var x,y,z: Integer;
  begin
  repeat
    Finde({%H-}x,{%H-}y,{%H-}z);
    if z>0 then SetZahl({%H-}x{%H-},y,{%H-}z,automatisch);
  until z=0;
  end;

var
  x,y,z: Integer;
begin
Result:=False;
EinfacheFinden;
if Unloesbar then
  begin
  Result:=False;
  Exit;
  end;
if Geloest then
  begin
  Result:=True;
  RemoveInternalUndos;
  Exit;
  end;
GetFreiesFeld({%H-}x,{%H-}y);
for z:=1 to 9 do
  begin
  if z in erlaubt[x,y] then
    begin
    AddUndo('x');
    SetZahl(x,y,z,geraten);
    if FindeLoesung
      then
        begin
        Result:=True;
        Exit;
        end
      else
        begin
        //Holzweg
        Undo;
        end;
    end;
  end;
end;

function TSudoku.Geloest: Boolean;
var
  x,y: Integer;
begin
//Alle Felder müssen belegt sein
Result:=False;
for y:=0 to 8 do for x:=0 to 8 do if Feld[x,y]=0 then Exit;
Result:=True;
end;

function TSudoku.GetFindGrund: String;
begin
Result:=grund;
end;

function TSudoku.getHint(x,y: Integer): String;
var
  i: Integer;
begin
Result:='';
for i:=1 to 9 do if i in erlaubt[x,y] then Result:=Result+IntToStr(i)+' ';
end;

function TSudoku.GetStatus(x,y: Integer): TFeldStatus;
begin
Result:=Status[x,y];
end;

function TSudoku.GetUndoString: String;
begin
if CanUndo
  then Result:=UndoList.Strings[UndoList.Count-1]
  else Result:='nichts';
end;

function TSudoku.GetZahl(x,y: Integer): Integer;
begin
Result:=Feld[x,y];
end;

procedure TSudoku.LoadFromFile(filename: String);
var
  f: TextFile;
  x,y,i: Integer;
  line, s: String;
begin
AssignFile(f,filename);
Reset(f);
for y:=0 to 8 do
  begin
  repeat
  ReadLn(f,line);
  s:='';
  for i:=1 to length(line) do
    begin
    if line[i] in ['1'..'9']
      then s:=s+line[i]
      else if line[i]=' ' then s:=s+'0';
    end;
  until length(s)=9;
  for x:=0 to 8 do
    begin
    Feld[x,y]:=StrToInt(s[x+1]);
    if Feld[x,y]=0
      then Status[x,y]:=manuell
      else Status[x,y]:=geladen;
    end;
  end;
CloseFile(f);
ErmittleAusschlussNeu;
end;

procedure TSudoku.LoescheZahl(x,y: Integer);
begin
Feld[x,y]:=0;
Status[x,y]:=manuell;
ErmittleAusschlussNeu;
end;

procedure TSudoku.SaveToFile(filename: String);
var
  f: TextFile;
  x,y: Integer;
begin
AssignFile(f,filename);
ReWrite(f);
for y:=0 to 8 do
  begin
  if y mod 3=0 then WriteLn(f,'+---+---+---+');
  for x:=0 to 8 do
    begin
    if x mod 3=0 then Write(f,'|');
    if Feld[x,y]=0
      then Write(f,' ')
      else Write(f,Feld[x,y]);
    end;
  WriteLn(f,'|');
  end;
WriteLn(f,'+---+---+---+');
CloseFile(f);
end;

procedure TSudoku.SetZahl(x,y,zahl: Integer;stat: TFeldStatus);
begin
Feld[x,y]:=zahl;
Status[x,y]:=stat;
SchliesseAusVon(x,y);
end;

procedure TSudoku.Undo;
var
  p: Pointer;
begin
  if UndoList.Count=0 then Exit;
  p:=UndoList.Objects[UndoList.Count-1];
  UndoList.Delete(UndoList.Count-1);
  Move(p^,Feld,SizeOf(Feld));
  Move((p+SizeOf(Feld))^,Status,SizeOf(Status));                                //  Move(Pointer(LongWord(p)+SizeOf(Feld))^,Status,SizeOf(Status));
  Move((p+SizeOf(Feld)+SizeOf(Status))^,Erlaubt,SizeOf(Erlaubt));               //  Move(Pointer(LongWord(p)+SizeOf(Feld)+SizeOf(Status))^,Erlaubt,SizeOf(Erlaubt));
  FreeMem(p);
end;

function TSudoku.Unloesbar: Boolean;
var
  x,y: Integer;
begin
//Simple Funktion: sucht Felder, auf die keine Zahl mehr darf
Result:=True;
for y:=0 to 8 do for x:=0 to 8 do if erlaubt[x,y]=[] then Exit;
Result:=False;
end;

procedure TSudoku.ErmittleAusschlussNeu;
var
  i,j: Integer;
begin
for i:=0 to 8 do for j:=0 to 8 do erlaubt[i,j]:=[1,2,3,4,5,6,7,8,9];
for i:=0 to 8 do for j:=0 to 8 do if Feld[i,j]>0 then SchliesseAusVon(i,j);
end;

procedure TSudoku.GetFreiesFeld(var x,y: Integer);
var
  i,j: Integer;
begin
for j:=0 to 8 do for i:=0 to 8 do if Feld[i,j]=0 then
  begin
  x:=i;
  y:=j;
  Exit;
  end;
end;

procedure TSudoku.RemoveInternalUndos;
var
  p: Pointer;
begin
while GetUndoString='x' do
  begin
  p:=UndoList.Objects[UndoList.Count-1];
  UndoList.Delete(UndoList.Count-1);
  FreeMem(p);
  end;
end;

procedure TSudoku.SchliesseAusVon(x,y: Integer);
var
  i, zahl: Integer;
begin
zahl:=Feld[x,y];
for i:=0 to 8 do
  begin
  // 1. Zeilen
  Exclude(erlaubt[x,i],zahl);
  // 2. Spalten
  Exclude(erlaubt[i,y],zahl);
  // 3x3-Felder
  Exclude(erlaubt[3*(x div 3)+i div 3,3*(y div 3)+i mod 3],zahl);
  end;
erlaubt[x,y]:=[Feld[x,y]];
end;
{%endregion TSudoku}
end.

